import unittest as ut

from tests._utils._timer import Timer
from drivelink import cached
#from Process import freeze_support


def uncachedFib(a):
    if a in [0, 1]:
        return a
    if a < 0:
        raise Exception("Reverse fibonacci sequence not implemented.")
    return uncachedFib(a - 1) + uncachedFib(a - 2)


def test_fib():
    assert uncachedFib(0) == 0
    assert uncachedFib(1) == 1
    assert uncachedFib(2) == 1
    assert uncachedFib(3) == 2
    assert uncachedFib(4) == 3
    assert uncachedFib(5) == 5


class cachedTest(ut.TestCase):
    c = None

    def setUp(self):
        @cached(self.id(), 1, 1)
        def fib(a):
            if a in [0, 1]:
                return a
            if a < 0:
                raise Exception("Reverse fibonacci sequence not implemented.")
            return fib(a - 1) + fib(a - 2)
        self.c = fib

    def test_fib(self):
        self.assertEqual(
            self.c(0), 0, "The zeroth element of the Fibonnaci sequence is 0, not {0}.".format(str(self.c(0))))
        self.assertEqual(
            self.c(1), 1, "The first element of the Fibonnaci sequence is 1, not {0}.".format(str(self.c(1))))
        self.assertEqual(
            self.c(2), 1, "The second element of the Fibonnaci sequence is 1, not {0}.".format(str(self.c(2))))
        self.assertEqual(
            self.c(3), 2, "The third element of the Fibonnaci sequence is 2, not {0}.".format(str(self.c(3))))
        self.assertEqual(
            self.c(4), 3, "The fourth element of the Fibonnaci sequence is 3, not {0}.".format(str(self.c(4))))
        self.assertEqual(
            self.c(5), 5, "The fifth element of the Fibonnaci sequence is 5, not {0}.".format(str(self.c(5))))

    def test_init(self):
        self.assertEqual(len(self.c.c), 0, "The cache was malformed.")
        self.assertEqual(self.c.f(0), uncachedFib(0), "The function was not entered correctly.")

    def test_cache(self):
        i = self.c(0)
        self.assertEqual(len(self.c.c), 1, "The value was not cached properly.")
        self.assertEqual(self.c(0), i, "The cached answer was incorrect.")

    def test_speed(self):
        with Timer() as t1:
            _ = uncachedFib(32)
        self.c.n = -1
        with Timer() as t2:
            _ = self.c(32)
        self.assertTrue(t2.interval < t1.interval,
                        "There isn't a speed up... This is useless then, I suppose.")
        with Timer() as t1:
            _ = self.c(32)
        self.assertTrue(t2.interval > t1.interval,
                        "There isn't a speed up... This is useless then, I suppose.")


if __name__ == '__main__':
    freeze_support()
    ut.main()



from collections import MutableMapping
from os.path import expanduser, join
from glob import glob

from drivelink import Link


class OrderedDict(Link, MutableMapping):
    """
    A dictionary class that maintains O(1) look up and write while keeping RAM usage O(1) as well.

    This is accomplished through a rudimentary (for now) hashing scheme to page the
    dictionary into parts.
    """

    def __init__(self, file_basename, size_limit=1024, max_pages=16, file_location=join(expanduser("~"), ".DriveLink"), compression_ratio=0):
        self.pages = {}
        self._total = set()
        super(OrderedDict, self).__init__(file_basename, size_limit, max_pages, file_location, compression_ratio)

    def copy_from(self, other):
        for key in other:
            self[key] = other[key]

    def load_index(self):
        other_values = super(OrderedDict, self).load_index()
        if other_values is None:
            return
        for f in glob(self._file_base + '*'):
            try:
                self._total.add(int(f[len(self._file_base):]))
            except ValueError:
                pass

    def open_page(self, k):
        if k in self._total:
            self._load_page_from_disk(k)
        if k not in self.pages:
            self.pages[k] = {}
            self._total.add(k)
            self._queue.append(k)

    def determine_index(self, key):
        """
        Figures out where the key in question should be.
        """
        # TODO: Refactor into b-tree
        k = hash(key) // self.size_limit
        return k, key

    def page_indices(self):
        for k in list(self._total):
            yield k

    def __contains__(self, item):
        try:
            i, k = self._finditem(item)
        except:
            return False
        return k in self.pages[i]

    def page_removed(self, number):
        self._total.remove(number)

    def __str__(self):
        return "Dictionary with values stored to " + self._file_base



"""
.. moduleauthor:: Chris Dusold <DriveLink@chrisdusold.com>

A module containing storage classes that maintain small RAM usage and original structure access order.

The motivation for this module was to provide constant size RAM usage
while maintaining normal use of Python Dictionaries and possibly other
structures for semi-big data, where it isn't large enough to warrant more
big data centric solutions.

"""
from drivelink._disklink import Link
from drivelink._diskdict import Dict
from drivelink._disklist import List
from drivelink._diskmemoize import cached
from drivelink._ordereddiskdict import OrderedDict



from collections import MutableSequence
from os.path import expanduser, join

from drivelink import Link


class List(Link, MutableSequence):
    """
    A list class that maintains O(k) look up and O(1) append while keeping RAM usage O(1) as well.
    Unfortunately, insert is O(n/k).

    This is accomplished through paging every size_limit consecutive values together
    behind the scenes.

    The object created can be used any way a normal list would be used, and will
    clean itself up on python closing. This means saving all the remaining pages
    to disk. If the file_basename and file_location was used before, it will load
    the old values back into itself so that the results can be reused.

    There are two ways to initialize this object, as a standard object:

        >>> diskList = List("samplelist")
        >>> for i in range(10):
        ...     diskList.append(i)
        ...
        >>> diskList[3]
        3
        >>> ", ".join(str(x) for x in diskList)
        '0, 1, 2, 3, 4, 5, 6, 7, 8, 9'
        >>> del diskList[5]
        >>> ", ".join(str(x) for x in diskList)
        '0, 1, 2, 3, 4, 6, 7, 8, 9'

    Or through context:

        >>> with List("testlist") as d:
        ...     for i in range(10):
        ...         d.append(i)
        ...     print(d[3])
        3

    If there is a way to break list like behavior and you can reproduce it, please
    report it to `the GitHub issues <https://github.com/cdusold/DriveLink/issues/>`_.
    """

    def __init__(self, file_basename, size_limit=1024, max_pages=16, file_location=join(expanduser("~"), ".DriveLink"), compression_ratio=0):
        self.pages = dict()
        self._number_of_pages = 0
        super(List, self).__init__(file_basename, size_limit, max_pages, file_location, compression_ratio)

    def copy_from(self, other):
        for value in other:
            self.append(value)

    def load_index(self):
        other_values = super(List, self).load_index()
        if other_values is None:
            return
        self._number_of_pages = other_values[0]

    def store_index(self):
        super(List, self).store_index(self._number_of_pages)

    def open_page(self, k):
        if 0 <= k < self._number_of_pages:
            self._load_page_from_disk(k)
        else:
            raise IndexError

    def determine_index(self, key):
        """
        Figures out where the key in question should be.
        """
        if key < 0:
            key += self._length
        return divmod(key, self.size_limit)

    def page_indices(self):
        for k in range(self._number_of_pages):
            yield k

    def __delitem__(self, key):
        '''
         Deletes the key value in question from the pages.
        '''
        super(List, self).__delitem__(key)
        i, _ = self.determine_index(key)
        for i in range(i, self._number_of_pages - 1):
            self.open_page(i + 1)
            if self.pages[i + 1]:
                v = self.pages[i + 1][0]
                del self.pages[i + 1][0]
                self.pages[i].append(v)
                self._guarantee_page(i + 1)
        self._guarantee_page(self._number_of_pages - 1)
        if not self.pages[self._number_of_pages - 1]:
            del self.pages[self._number_of_pages - 1]
            self._number_of_pages -= 1

    def __reversed__(self):
        for p in reversed(range(self._number_of_pages)):
            self._guarantee_page(p)
            for i in reversed(self.pages[p]):
                yield i

    def page_removed(self, number):
        self._number_of_pages -= 1

    def __str__(self):
        return "List with values stored to " + self._file_base

    def append(self, v):
        k = self._length // self.size_limit
        if k == self._number_of_pages:
            self._newpage()
        self._guarantee_page(k)
        self.pages[k].append(v)
        self._length += 1

    def insert(self, i, v):
        k, i = divmod(i, self.size_limit)
        if k == self._number_of_pages:
            self._newpage()
        self._guarantee_page(k)
        self.pages[k].insert(i, v)
        if len(self.pages[k]) > self.size_limit:
            for k in range(k, self._number_of_pages - 1):
                self._guarantee_page(k)
                v = self.pages[k][-1]
                del self.pages[k][-1]
                self._guarantee_page(k + 1)
                self.pages[k + 1].insert(0, v)
            if len(self.pages[self._number_of_pages - 1]) > self.size_limit:
                self._newpage()
                self.pages[self._number_of_pages -
                           1].append(self.pages[self._number_of_pages - 2][-1])
                del self.pages[self._number_of_pages - 2][-1]
        self._length += 1

    def _newpage(self):
        self.pages[self._number_of_pages] = []
        self._queue.append(self._number_of_pages)
        self._number_of_pages += 1



import copy
from os.path import join, expanduser

from drivelink import Dict


class _Cached:
    "This will shine the most with recursive functions. But the recursion has to call the cached function, not the function itself."
    f = None
    c = None

    def __init__(self, function, file_basename=None, size_limit=1024, max_pages=16, file_location=join(expanduser("~"), ".DriveLink"), compression_ratio=0):
        for n in list(n for n in set(dir(function)) - set(dir(self)) if n != '__class__'):
            setattr(self, n, getattr(function, n))
        if file_basename is None:
            file_basename = function.__name__
        self.f = function
        self.c = Dict(file_basename, size_limit, max_pages, file_location, compression_ratio)

    def __call__(self, *args, **kwargs):
        i = str(args) + str(kwargs)
        if i in self.c:
            return copy.deepcopy(self.c[i])
        else:
            t = self.f(*args, **kwargs)
            self.c[i] = copy.deepcopy(t)
            return t


def cached(file_basename=None, size_limit=1024, max_pages=16, file_location=join(expanduser("~"), ".DriveLink")):
    '''
    A decorator that creates a simplistic cached function with minimal overhead.

    This provides very simplistic and quick cache. The values are saved to a drivelink.Dict
    and will be reloaded on program restarting.
    '''
    def decorator(f):
        return _Cached(f, file_basename, size_limit, max_pages, file_location)
    return decorator



from drivelink import List
import pytest
import os
#from Process import freeze_support


def test_list():
    lst = List("testList")
    for i in range(10):
        lst.append(i)
    for i in range(10):
        assert lst[i] == i


def test_change_settings():
    with List("testListChangeSettings", 1, 1) as l:
        l.append(1)
        l.append("c")
        l.append(3.4)
    with List("testListChangeSettings", 2, 2) as l:
        assert l[0] == 1
        assert l[1] == "c"
        assert l[2] == 3.4
    with List("testListChangeSettings", 1, 1) as l:
        assert l[0] == 1
        assert l[1] == "c"
        assert l[2] == 3.4


def test_init():
    with pytest.raises(ValueError) as excinfo:
        List("testListInit", 0)
    excinfo.match(".* per page.*")
    with pytest.raises(ValueError) as excinfo:
        List("testListInit", 1, 0)
    excinfo.match(".* in RAM.*")
    with List("testListInit", 1, 1):
        pass
    List("testListInit", 1, 1)


def test_guarantee_page():
    with List("testListGuaranteePage", 1, 1) as l:
        l.append(1)
        l.append("c")
        l.append(3.4)
        assert l[0] == 1
        assert l[1] == "c"
        assert l[2] == 3.4


def test_save():
    l = List("testListSave", 1, 1)
    l.append(1)
    l.append("c")
    l.append(3.4)
    l.close()
    with List("testListSave", 1, 1) as l:
        assert l[0] == 1
        assert l[1] == "c"
        assert l[2] == 3.4


def test_string_funcs():
    l = List("testListStringFuncs")
    assert str(l).startswith("List ")
    assert str(l).endswith("testListStringFuncs")
    assert repr(l).startswith("List(")
    assert repr(l).endswith(".DriveLink')")


def test_insert():
    l = List("testInsert", 1, 1)
    l.append(0)
    l.extend([1, 3])
    l.insert(2, 2)
    for i in range(4):
        assert l[i] == i
    l = List("testInsert", 2, 2)
    l.insert(1,-1)
    assert l[1] == -1
    assert l[2] == 1


def test_delete():
    l = List("testDelete", 1, 1)
    l.append(0)
    l.extend([0, 1, 2, 5, 3])
    del l[1]
    del l[3]
    for i in range(4):
        assert l[i] == i


def test_reverse():
    l = List("testReverse", 2, 2)
    l.extend([i for i in reversed(range(10))])
    for i, v in enumerate(reversed(l)):
        assert i == v
    for i in range(1,11):
        assert l[-i] == i-1

def test_contains():
    with List("testListContains", 1, 1) as l:
        l.append(1)
        l.append("c")
        l.append(3.4)
        assert 1 in l
        assert "c" in l
        assert 3.4 in l
        assert 2 not in l
        assert "d" not in l
        assert 3.3 not in l


if __name__ == '__main__':
    freeze_support()
    ut.main()



from drivelink import OrderedDict
import pytest
import os
#from Process import freeze_support


def test_dict():
    dct = OrderedDict("testOrderedDict")
    for i in range(10):
        dct[i] = i
    for i in range(10):
        assert dct[i] == i


def test_init():
    with pytest.raises(ValueError) as excinfo:
        OrderedDict("testOrderedDictInit", 0)
    excinfo.match(".* per page.*")
    with pytest.raises(ValueError) as excinfo:
        OrderedDict("testOrderedDictInit", 1, 0)
    excinfo.match(".* in RAM.*")
    with OrderedDict("testOrderedDictInit", 1, 1):
        pass
    OrderedDict("testOrderedDictInit", 1, 1)


def test_change_settings():
    with OrderedDict("testOrderedDictChangeSettings", 1, 1) as d:
        d[0] = 1
        d[1] = "c"
        d[2] = 3.4
    with OrderedDict("testOrderedDictChangeSettings", 2, 2) as d:
        assert d[0] == 1
        assert d[1] == "c"
        assert d[2] == 3.4
    with OrderedDict("testOrderedDictChangeSettings", 1, 1) as d:
        assert d[0] == 1
        assert d[1] == "c"
        assert d[2] == 3.4


def test_guarantee_page():
    with OrderedDict("testOrderedDictGuaranteePage", 1, 1) as d:
        d[0] = 1
        d[1] = "c"
        d[2] = 3.4
        assert d[0] == 1
        assert d[1] == "c"
        assert d[2] == 3.4


def test_string_funcs():
    d = OrderedDict("testOrderedDictStringFuncs")
    assert str(d).startswith("Dictionary")
    assert str(d).endswith("testOrderedDictStringFuncs")
    assert repr(d).startswith("OrderedDict(")
    assert repr(d).endswith(".DriveLink')")


def test_save():
    d = OrderedDict("testOrderedDictSave", 1, 1)
    d[0] = 1
    d[1] = "c"
    d[2] = 3.4
    d.close()
    with OrderedDict("testOrderedDictSave", 1, 1) as d:
        assert d[0] == 1
        assert d[1] == "c"
        assert d[2] == 3.4

def test_contains():
    with OrderedDict("testOrderedDictContains", 1, 1) as d:
        d[0] = 1
        d[1] = "c"
        d[2] = 3.4
        assert 0 in d
        assert 1 in d
        assert 2 in d
        assert 3 not in d
        assert 4 not in d
        assert 5 not in d


if __name__ == '__main__':
    freeze_support()
    ut.main()



import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="DriveLink",
    version="0.2.1.1",
    author="Chris Dusold",
    author_email="DriveLink@ChrisDusold.com",
    description=("A set of memory conserving data structures."),
    license=read("LICENSE"),
    keywords="memory",
    url="http://drivelink.rtfd.org/",
    packages=['drivelink', 'tests'],
    long_description=read('README.rst'),
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",  # Hopefully.
        "Topic :: Scientific/Engineering",
        "Topic :: Utilities",
    ],
)



try:
    import cPickle as pickle
except:
    import pickle
from os.path import expanduser, join, split
from os import remove, makedirs, rename
from glob import glob
import atexit
import zlib


class Link(object):
    """
    This abstract base class provides shared functionality for any hard disk linked
    class required. The other classes in this library use this class, and can be
    referred to if you need to implement your own version. (Please consider a pull
    request at https://github.com/cdusold/DriveLink if you make a good general
    implementation.)

    .. attention:: All classes in DriveLink use this class, so the following applies
                   to each of them.

    To be able to implement your own, in addition to implementing the abstract functions,
    you have to implement self.pages as a dictionary that will work for your class.

    This base class provides wrapping that automatically saves to disk, if everything
    is implemented correctly in inheriting classes. It provides the ability to access
    implementing classes directly (direct use of Class.close() reccommended) or
    through a context manager.

    .. note:: This abstract class is not thread safe, nor is it process safe. Any multithreaded
              or multiprocessed uses of implemented classes hold no guarantees of accuracy.

    You can configure how this class stores things in a few ways.

    The file_basename parameter allows you to keep multiple different stored objects
    in the same file_location, which defaults to .DriveLink in the user's home folder.
    Using a file_basename of the empty string may cause a small slowdown if more
    than just this object's files are in the folder. Using substrings of other basenames
    or basenames that end in numbers may cause irregular behavior. Using a file_location
    of the empty string will result in files being placed in the environment's current
    location (i.e. what `os.getcwd()` would return).

    The size_limit parameter determines how many items are kept in each page, and the
    max_pages parameter determines how many pages can be kept in memory at the same
    time. If you use smaller items in the class, increasing either is probably a
    good idea to get better performance. This setting will only use about 128 MB if
    standard floats or int32 values. Likely less than 200 MB will ever be in memory,
    which prevents the RAM from filling up and needing to use swap space. Tuning
    these values will be project, hardware and usage specific to get the best results.
    Even with the somewhat low defaults, this will beat out relying on python to
    use swap space.

    In order to speed up disk access, you can specify a compression_ratio. compression
    is performed using Python's built in `ZLib library <https://docs.python.org/library/zlib.html>`_.
    """

    def __init__(self, file_basename, size_limit=1024, max_pages=16, file_location=join(expanduser("~"), ".DriveLink"), compression_ratio=0):
        if max_pages < 1:
            raise ValueError("There must be allowed at least one page in RAM.")
        self.max_pages = max_pages
        if size_limit < 1:
            raise ValueError("There must be allowed at least one item per page.")
        self.size_limit = size_limit
        if file_location:
            try:
                makedirs(file_location)
            except OSError as e:
                if e.errno != 17:
                    raise
                pass
        self._file_base = join(file_location, file_basename)
        self._file_loc = file_location
        self._file_basename = file_basename
        self._compression = compression_ratio
        self._length = 0
        self._queue = []
        # Just in case, cache pickle.
        self._pickle = pickle
        self._check_old_settings()
        self.load_index()
        atexit.register(Link.close, self)

    def _check_old_settings(self):
        """
        This loads and saves the settings used to create the drivelink. If different
        settings were used originally, this may conflict with the operation of the
        link, so the values will be copied out into the new structure.
        """
        try:
            with open(self._file_base + 'Set', 'rb') as f:
                size_limit = self._pickle.load(f)
            if size_limit != self.size_limit:
                self._make_old_values_available(size_limit)
        except IOError:
            pass
        with open(self._file_base + 'Set', 'wb') as f:
             self._pickle.dump(self.size_limit, f)

    def _make_old_values_available(self, size_limit):
        """
        In order to take advantage of lazy loading, it may be worth your time to
        overload this function when you inherit and implement a recovery method yourself.

        You have to implement the copy_from function to be able to use this anyway.
        """
        for file_name in glob(self._file_base+"*"):
            path, name = split(file_name)
            rename(file_name, join(path, "~"+name))
        with type(self)(self._file_base, self.size_limit, min(4, self.max_pages), self._file_loc) as new:
            with type(self)("~"+self._file_basename, size_limit, 1, self._file_loc, -1) as old:
                new.copy_from(old)
        for file_name in glob(join(self._file_loc, "~"+self._file_basename+"*")):
            remove(file_name)

    def copy_from(self, other):
        """
        Depending on the type of container this link is wrapping, this will transfer
        values from a container of the same time to the new one.
        """
        raise NotImplementedError

    _stored_index = None
    def load_index(self):
        """
        This base implementation loads the number of entries in the collection and
        returns anything else that may be stored with it. If the inheriting class
        requires more than just a length value to index its items, it is reccommended
        to override this method and store_index, so that loading is automatic.
        """
        try:
            with open(self._file_base + 'Len', 'rb') as f:
                self._stored_index = f.read()
                other_values = self._pickle.loads(self._stored_index)
                other_values, self._length = other_values[:-1], other_values[-1]
        except IOError:
            return None
        return other_values

    def store_index(self, *other_values):
        """
        This base implementation saves the number of entries in the collection and
        anything passed into the function with it. If the inheriting class
        requires more than just a length value to index its items, it is reccommended
        to override this method and load_index, so that loading is automatic.

        To save additional items, just pass them as arguments to a super call.
        """
        to_save = self._pickle.dumps(tuple(other_values) + (self._length,))
        if to_save == self._stored_index:
            return
        with open(self._file_base + 'Len', 'wb') as f:
            f.write(to_save)
        self._stored_index = to_save

    def __len__(self):
        '''
        Returns the number of entries stored.
        '''
        return self._length

    def __repr__(self):
        return (self.__class__.__name__ + "('" + self._file_basename + "', " +
                str(self.size_limit) + ', ' + str(self.max_pages) + ", '" + self._file_loc + "')")

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_val, trace):
        self.close()
        del self

    def close(self):
        '''
        Save all the values to disk before closing.
        '''
        if (self is None or not hasattr(self, "_save_page_to_disk")
                or not hasattr(self, "_file_base") or self._file_base is None):
            return
        while len(self.pages) > 0:
            for key in set(self.pages.keys()):
                self._save_page_to_disk(key)

    def _guarantee_page(self, k):
        """
        Ensures the page is available.
        """
        if k not in self.pages:
            self.open_page(k)
        while len(self._queue) > self.max_pages:
            if self._queue[0] == k:
                self._queue.append(self._queue[0])
                del self._queue[0]
            self._save_page_to_disk(self._queue[0])

    def open_page(self, k):
        """
        Once it is determined a page isn't in RAM, make it so.

        This will load a page if it is saved to disk, or create a new one when necessary.
        """
        raise NotImplementedError

    def _finditem(self, key):
        """
        Pulls up the page containing the key in question.
        """
        k, i = self.determine_index(key)
        self._guarantee_page(k)
        return k, i

    def determine_index(self, key):
        """
        Figures out where the key in question should be.
        """
        raise NotImplementedError

    def __setitem__(self, key, value):
        '''
         Sets a value that a key maps to.
        '''
        i, k = self._finditem(key)
        if k not in self.pages[i]:
            # If this isn't possible, the implemented page should throw an error.
            self.pages[i][k] = value
            self._length += 1
        else:
            self.pages[i][k] = value

    def __getitem__(self, key):
        '''
         Retrieves the value the key maps to.
        '''
        i, k = self._finditem(key)
        return self.pages[i][k]

    def __delitem__(self, key):
        '''
         Deletes the entry in question from the pages.
        '''
        i, k = self._finditem(key)
        del self.pages[i][k]
        self._length -= 1

    def __contains__(self, item):
        for page in self._iterpages():
            if item in page:
                return True
        return False

    def _iterpages(self):
        """
        Pulls up page after page and cycles through all of them.
        """
        for k in self.page_indices():
            self._guarantee_page(k)
            yield self.pages[k]

    def __iter__(self):
        '''
         Iterates through all the keys stored.
        '''
        for p in self._iterpages():
            for i in p:
                yield i

    def page_indices(self):
        """
        Yields an iterable containing each page.
        """
        raise NotImplementedError

    def _save_page_to_disk(self, number):
        self.store_index()
        if self._file_base:
            if number in self.pages:
                if len(self.pages[number]) > 0:
                    to_save = self._pickle.dumps(self.pages[number])
                    if self._compression:
                        to_save = zlib.compress(to_save, self._compression)
                    with open(self._file_base + str(number), 'wb') as f:
                        f.write(to_save)
                else:
                    self.page_removed(number)
                del self.pages[number]
            for i in range(len(self._queue)):
                if self._queue[i] == number:
                    del self._queue[i]
                    break

    def _load_page_from_disk(self, number):
        if self._file_base:
            with open(self._file_base + str(number), 'rb') as f:
                to_load = f.read()
            try:
                to_load = zlib.decompress(to_load)
            except zlib.error:
                pass
            self.pages[number] = self._pickle.loads(to_load)
            self._queue.append(number)
            remove(self._file_base + str(number))



from collections import MutableMapping
from os.path import expanduser, join
from glob import glob

from drivelink import Link


class _page(dict):
    currentDepth = 0


class Dict(Link, MutableMapping):
    """
    A dictionary class that maintains O(1) look up and write while keeping RAM usage O(1) as well.

    This is accomplished through a rudimentary (for now) hashing scheme to page the
    dictionary into parts.

    The object created can be used any way a normal dict would be used, and will
    clean itself up on python closing. This means saving all the remaining pages
    to disk. If the file_basename and file_location was used before, it will load
    the old values back into itself so that the results can be reused.

    There are two ways to initialize this object, as a standard object:

        >>> diskDict = Dict("sampledict")
        >>> for i in range(10):
        ...     diskDict[i] = chr(97+i)
        ...
        >>> diskDict[3]
        'd'
        >>> 5 in diskDict
        True
        >>> del diskDict[5]
        >>> ", ".join(str(x) for x in diskDict.keys())
        '0, 1, 2, 3, 4, 6, 7, 8, 9'
        >>> 5 in diskDict
        False

    Or through context:

        >>> with Dict("testdict") as d:
        ...     for i in range(10):
        ...         d[i] = chr(97+i)
        ...     print(d[3])
        d

    If there is a way to break dict like behavior and you can reproduce it, please
    report it to `the GitHub issues <https://github.com/cdusold/DriveLink/issues/>`_.
    """

    def __init__(self, file_basename, size_limit=1024, max_pages=16, file_location=join(expanduser("~"), ".DriveLink"), compression_ratio=0):
        self.pages = _page()
        self._total = set()
        super(Dict, self).__init__(file_basename, size_limit, max_pages, file_location, compression_ratio)

    def copy_from(self, other):
        for key in other:
            self[key] = other[key]

    def load_index(self):
        other_values = super(Dict, self).load_index()
        if other_values is None:
            return
        self.pages.currentDepth = other_values[0]
        for f in glob(self._file_base + '*'):
            try:
                self._total.add(int(f[len(self._file_base):]))
            except ValueError:
                pass

    def store_index(self):
        super(Dict, self).store_index(self.pages.currentDepth)

    def open_page(self, k):
        if k in self._total:
            self._load_page_from_disk(k)
        if k not in self.pages:
            self.pages[k] = _page()
            self._total.add(k)
            self.pages[k].currentDepth = self.pages.currentDepth
            self._queue.append(k)

    def determine_index(self, key):
        """
        Figures out where the key in question should be.

        Most frequently O(1), when a page becomes too large, there's
        a O(ln(n)) search that refactors O(k ln(n)) elements
        once every O(k) insertions. A full refactor usually
        happens in strides, moving a total of O(n) elements
        split up over O(ln(n)) calls. This makes the worst
        time refactor O(n) and usual refactor approximately
        O(n/ ln(n)). Average case lookup O(n/k).
        """
        k = hash(key) & self.pages.currentDepth
        i = 0
        while (k & (self.pages.currentDepth >> i)) not in self._total | set([0]):
            i += 1
        self._branchpage(k & (self.pages.currentDepth >> i))
        return k, key

    def page_indices(self):
        for k in list(self._total):
            yield k

    def __setitem__(self, key, value):
        '''
         Sets a value that a key maps to.
        '''
        super(Dict, self).__setitem__(key, value)
        i, _ = self.determine_index(key)
        if len(self.pages[i]) > self.size_limit:
            if self.pages[i].currentDepth == self.pages.currentDepth:
                self.pages.currentDepth <<= 1
                self.pages.currentDepth |= 1
            self._branchpage(i)

    def __contains__(self, item):
        try:
            i, k = self._finditem(item)
        except:
            return False
        return k in self.pages[i]

    def page_removed(self, number):
        self._total.remove(number)

    def __str__(self):
        return "Dictionary with values stored to " + self._file_base

    def _branchpage(self, pagenumber):
        self._guarantee_page(pagenumber)
        if self.pages[pagenumber].currentDepth == self.pages.currentDepth:
            return
        self.pages[pagenumber].currentDepth = self.pages.currentDepth
        for key in set(self.pages[pagenumber].keys()):
            k = hash(key) & self.pages.currentDepth
            if k != pagenumber:
                self._guarantee_page(pagenumber)
                v = self.pages[pagenumber][key]
                del self.pages[pagenumber][key]
                self._guarantee_page(k)
                self.pages[k][key] = v
